/*	$NetBSD: adt7467.c,v 1.13 2008/05/08 02:03:22 macallan Exp $	*/

/*-
 * Copyright (C) 2008 The NetBSD Foundation
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * a driver for the dBCOOL(tm) family of environmental controllers
 *
 * Data sheets for the various supported chips are available at
 * http://www.onsemi.com
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/device.h>
#include <sys/malloc.h>
#include <sys/sysctl.h>

#include <uvm/uvm_extern.h>

#include <dev/i2c/dbcool_var.h>
#include <dev/i2c/dbcool_reg.h>

static void dbcool_attach(device_t, device_t, void *);
static int dbcool_match(device_t, cfdata_t, void *);
static int dbcool_chip_ident(struct dbcool_softc *);
static uint8_t dbcool_readreg(struct dbcool_softc *, uint8_t);
static void dbcool_writereg(struct dbcool_softc *, uint8_t, uint8_t);
static void dbcool_refresh(struct sysmon_envsys *, envsys_data_t *);
static void dbcool_setup(device_t self);

uint8_t dbcool_tempoffset(struct dbcool_softc *);
int dbcool_readtemp(struct dbcool_softc *, uint8_t);
int dbcool_readrpm(struct dbcool_softc *, uint8_t);
int dbcool_supply_voltage(struct dbcool_softc *);
int dbcool_read_voltage(struct dbcool_softc *, uint8_t);

struct dbcool_sensor ADT7475_sensor_table[] = {
	{ DBC_TEMP, 0, DBCOOL_LOCAL_TEMP,	"local temp" },
	{ DBC_TEMP, 1, DBCOOL_REMOTE1_TEMP,	"remote1 temp" },
	{ DBC_TEMP, 2, DBCOOL_REMOTE2_TEMP,	"remote2 temp" },
	{ DBC_VOLT, 0, DBCOOL_CPU_VOLTAGE,	"Vccp" },
	{ DBC_VOLT, 1, DBCOOL_SUPPLY_VOLTAGE,	"Vcc" },
	{ DBC_FAN,  0, DBCOOL_FAN1_TACH_LOW,	"fan1" },
	{ DBC_FAN,  1, DBCOOL_FAN2_TACH_LOW,	"fan2" },
	{ DBC_FAN,  2, DBCOOL_FAN3_TACH_LOW,	"fan3" },
	{ DBC_FAN,  3, DBCOOL_FAN4_TACH_LOW,	"fan4" },
	{ DBC_CTL,  0, DBCOOL_LOCAL_TMIN,	"local_Tmin" },
	{ DBC_CTL,  1, DBCOOL_REMOTE1_TMIN,	"remote1_Tmin" },
	{ DBC_CTL,  2, DBCOOL_REMOTE2_TMIN,	"remote2_Tmin" },
	{ DBC_CTL,  0, DBCOOL_LOCAL_TTHRESH,	"local_Tthresh" },
	{ DBC_CTL,  1, DBCOOL_REMOTE1_TTHRESH,	"remote1_Tthresh" },
	{ DBC_CTL,  2, DBCOOL_REMOTE2_TTHRESH,	"remote2_Tthresh" },
	{ 0, 0, 0, NULL }
};

struct dbcool_power_control ADT7475_power_table[] = {
	{ DBCOOL_PWM1_CTL, DBCOOL_PWM1_MINDUTY, DBCOOL_PWM1_MAXDUTY,
		DBCOOL_PWM1_CURDUTY, DBCOOL_PWM1_TRANGE,
		"fan_control_1" },
	{ DBCOOL_PWM2_CTL, DBCOOL_PWM2_MINDUTY, DBCOOL_PWM2_MAXDUTY,
		DBCOOL_PWM2_CURDUTY, DBCOOL_PWM2_TRANGE,
		"fan_control_2" },
	{ DBCOOL_PWM3_CTL, DBCOOL_PWM3_MINDUTY, DBCOOL_PWM3_MAXDUTY,
		DBCOOL_PWM3_CURDUTY, DBCOOL_PWM3_TRANGE,
		"fan_control_3" },
	{ 0, 0, 0, 0, 0, NULL }
};

struct dbcool_sensor ADT7466_sensor_table[] = {
	{ DBC_TEMP, 0, DBCOOL_LOCAL_TEMP_MSB,	"local temp" },
	{ DBC_TEMP, 1, DBCOOL_REMOTE_TEMP_MSB,	"remote temp" },
	{ DBC_VOLT, 0, DBCOOL_VCCALT,		"supply voltage" },
	{ DBC_VOLT, 1, DBCOOL_AIN1,		"analog V1" },
	{ DBC_VOLT, 2, DBCOOL_AIN2,		"analog V2" },
	{ DBC_FAN,  0, DBCOOL_FANA_LSB,		"fan1" },
	{ DBC_FAN,  1, DBCOOL_FANB_LSB,		"fan2" },
	{ 0, 0, 0, NULL }
};

struct dbcool_sensor ADM1027_sensor_table[] = {
	{ DBC_TEMP, 0, DBCOOL_LOCAL_TEMP,	"local temp" },
	{ DBC_TEMP, 1, DBCOOL_REMOTE1_TEMP,	"remote1 temp" },
	{ DBC_TEMP, 2, DBCOOL_REMOTE2_TEMP,	"remote2 temp" },
	{ DBC_VOLT, 0, DBCOOL_CPU_VOLTAGE,	"Vccp" },
	{ DBC_VOLT, 1, DBCOOL_SUPPLY_VOLTAGE,	"Vcc" },
	{ DBC_VOLT, 2, DBCOOL_25VIN,		"V2.5" },
	{ DBC_VOLT, 3, DBCOOL_5VIN,		"V5" },
	{ DBC_VOLT, 4, DBCOOL_12VIN,		"V12" },
	{ DBC_FAN,  0, DBCOOL_FAN1_TACH_LOW,	"fan1" },
	{ DBC_FAN,  1, DBCOOL_FAN2_TACH_LOW,	"fan2" },
	{ DBC_FAN,  2, DBCOOL_FAN3_TACH_LOW,	"fan3" },
	{ DBC_FAN,  3, DBCOOL_FAN4_TACH_LOW,	"fan4" },
	{ DBC_CTL,  0, DBCOOL_LOCAL_TMIN,	"local_Tmin" },
	{ DBC_CTL,  1, DBCOOL_REMOTE1_TMIN,	"remote1_Tmin" },
	{ DBC_CTL,  2, DBCOOL_REMOTE2_TMIN,	"remote2_Tmin" },
	{ DBC_CTL,  0, DBCOOL_LOCAL_TTHRESH,	"local_Tthresh" },
	{ DBC_CTL,  1, DBCOOL_REMOTE1_TTHRESH,	"remote1_Tthresh" },
	{ DBC_CTL,  2, DBCOOL_REMOTE2_TTHRESH,	"remote2_Tthresh" },
	{ 0, 0, 0, NULL }
};

struct chip_id chip_table[] = {
	{ DBCOOL_COMPANYID, ADT7476_DEVICEID, 0xff,
		ADT7475_sensor_table, ADT7475_power_table,
		DBCFLAG_TEMPOFFSET | DBCFLAG_HAS_PWMMUX | DBCFLAG_HAS_MAXDUTY,
		"ADT7476" },
	{ DBCOOL_COMPANYID, ADT7475_DEVICEID, 0xff,
		ADT7475_sensor_table, ADT7475_power_table,
		DBCFLAG_TEMPOFFSET | DBCFLAG_HAS_PWMMUX | DBCFLAG_HAS_MAXDUTY,
		"ADT7475" },
	{ DBCOOL_COMPANYID, ADT7473_DEVICEID, ADT7473_REV_ID,
		ADT7475_sensor_table, ADT7475_power_table,
		DBCFLAG_TEMPOFFSET | DBCFLAG_HAS_PWMMUX | DBCFLAG_HAS_MAXDUTY,
		"ADT7463" },
	{ DBCOOL_COMPANYID, ADT7473_DEVICEID, ADT7473_1_REV_ID,
		ADT7475_sensor_table, ADT7475_power_table,
		DBCFLAG_TEMPOFFSET | DBCFLAG_HAS_PWMMUX | DBCFLAG_HAS_MAXDUTY,
		"ADT7463-1" },
	{ DBCOOL_COMPANYID, ADT7467_DEVICEID, ADT7467_REV_ID1,
		ADT7475_sensor_table, ADT7475_power_table,
		DBCFLAG_TEMPOFFSET | DBCFLAG_HAS_PWMMUX | DBCFLAG_MULTI_VCC |
		DBCFLAG_HAS_MAXDUTY,
		"ADT7467/ADT7468" },
	{ DBCOOL_COMPANYID, ADT7467_DEVICEID, ADT7467_REV_ID2,
		ADT7475_sensor_table, ADT7475_power_table,
		DBCFLAG_TEMPOFFSET | DBCFLAG_HAS_PWMMUX | DBCFLAG_MULTI_VCC |
		DBCFLAG_HAS_MAXDUTY,
		"ADT7467/ADT7468" },
	{ DBCOOL_COMPANYID, ADT7466_DEVICEID, 0xff,
		ADT7466_sensor_table, NULL,
		DBCFLAG_MULTI_VCC2 | DBCFLAG_TEMPOFFSET,
		"ADT7466" },
	{ DBCOOL_COMPANYID, ADT7463_DEVICEID, ADT7463_REV_ID1,
		ADM1027_sensor_table, ADT7475_power_table,
		DBCFLAG_MULTI_VCC,
		"ADT7463" },
	{ DBCOOL_COMPANYID, ADT7463_DEVICEID, ADT7463_REV_ID2,
		ADM1027_sensor_table, ADT7475_power_table,
		DBCFLAG_MULTI_VCC,
		"ADT7463" },
	{ DBCOOL_COMPANYID, ADM1027_DEVICEID, ADM1027_REV_ID,
		ADM1027_sensor_table, ADT7475_power_table,
		DBCFLAG_MULTI_VCC,
		"ADM1027" },
	{ 0, 0, 0, NULL, NULL, 0, NULL }
};

CFATTACH_DECL_NEW(dbcool, sizeof(struct dbcool_softc),
    dbcool_match, dbcool_attach, NULL, NULL);

int
dbcool_match(device_t parent, cfdata_t cf, void *aux)
{
	struct i2c_attach_args *ia = aux;
	struct dbcool_softc sc;
	sc.sc_tag = ia->ia_tag;
	sc.sc_addr = ia->ia_addr;

	/* no probing if we attach to iic, but verify chip id */
	if (dbcool_chip_ident(&sc) >= 0)
		return 1;

	return 0;
}

void
dbcool_attach(device_t parent, device_t self, void *aux)
{
	struct dbcool_softc *sc = device_private(self);
	struct i2c_attach_args *args = aux;
	int i;

	sc->sc_addr = args->ia_addr;
	sc->sc_tag = args->ia_tag;
	i = dbcool_chip_ident(sc);
	sc->sc_chip = &chip_table[i];

	aprint_normal(" dBCool(tm) Controller\n");
	aprint_normal_dev(self, " %s rev %d\n", sc->sc_chip->name,
		      dbcool_readreg(sc, DBCOOL_REVISION_REG) & 0x0f);

	dbcool_setup(self);
}

uint8_t dbcool_readreg(struct dbcool_softc *sc, uint8_t reg)
{
	uint8_t data = 0;
	
	iic_acquire_bus(sc->sc_tag, 0);
	iic_exec(sc->sc_tag, I2C_OP_READ, sc->sc_addr, &reg, 1,
	    &data, 1, 0);
	iic_release_bus(sc->sc_tag, 0);
	return data;
}

void dbcool_writereg(struct dbcool_softc *sc, uint8_t reg, uint8_t data)
{
	uint8_t mdata[2] = {reg, data};
	
	iic_acquire_bus(sc->sc_tag, 0);
	iic_exec(sc->sc_tag, I2C_OP_WRITE, sc->sc_addr, &mdata, 2, NULL, 0, 0);
	iic_release_bus(sc->sc_tag, 0);
}

#if NSYSMON_ENVSYS > 0

/*
 * Some chips are capable of reporting an extended temperature range
 * by default.  On these models, config register 5 bit 0 can be set
 * to 1 for compatability with other chips that report 2s complement.
 */
uint8_t
dbcool_tempoffset(struct dbcool_softc *sc)
{
	uint8_t	cfg5;

	if (sc->sc_chip->flags & DBCFLAG_TEMPOFFSET) {
		cfg5 = dbcool_readreg(sc, DBCOOL_CONFIG5_REG);
		if (cfg5 & DBCOOL_CFG5_TWOSCOMP)
			return 0;
		else
			return 64;
	} else
		return 0;
}

int
dbcool_readtemp(struct dbcool_softc *sc, uint8_t reg)
{
	uint8_t	t1, t2, t3, val, ext;
	uint8_t offset = 0;
	int temp;

	if (reg == DBCOOL_LOCAL_TEMP_MSB || reg == DBCOOL_REMOTE_TEMP_MSB) {
		/* ADT7466 temps are in strange location */
		val = dbcool_readreg(sc, reg);
		ext = dbcool_readreg(sc, DBCOOL_CONFIG1A_REG);
		/* Also, offset-mode is controlled different in ADT7466 */
		if (ext & 0x80)
			offset = 64;
		ext = dbcool_readreg(sc, reg + 1);
	} else {
		ext = dbcool_readreg(sc, DBCOOL_EXTRES2_REG);

		/* Read all msb regs to unlatch them */
		t1 = dbcool_readreg(sc, DBCOOL_12VIN);
		t1 = dbcool_readreg(sc, DBCOOL_REMOTE1_TEMP);
		t2 = dbcool_readreg(sc, DBCOOL_REMOTE2_TEMP);
		t3 = dbcool_readreg(sc, DBCOOL_LOCAL_TEMP);
		switch (reg) {
		case DBCOOL_REMOTE1_TEMP:
			val = t1;
			ext >>= 2;
			break;
		case DBCOOL_LOCAL_TEMP:
			val = t3;
			ext >>= 4;
			break;
		case DBCOOL_REMOTE2_TEMP:
			val = t2;
			ext >>= 6;
			break;
		default:
			val = 0;
			break;
		}
		ext &= 0x03;

		offset = dbcool_tempoffset(sc);
	}

	/* Convert degC to uK and include extended precision bits */
	if (offset == 0)
		temp = (int8_t)val;
	else {
		temp =  val;
		temp -= offset;
	}
	temp *= 1000000;
	temp +=  250000 * (int)ext;
	temp += 273150000U;

	return temp;
}

int
dbcool_readrpm(struct dbcool_softc *sc, uint8_t reg)
{
	int rpm;
	uint8_t rpm_lo, rpm_hi;

	rpm_lo = dbcool_readreg(sc, reg);
	rpm_hi = dbcool_readreg(sc, reg + 1);

	rpm = (rpm_hi << 8) | rpm_lo;
	if (rpm == 0xffff)
		return -1;	/* indicates stalled/failed fan */

	return (90000 * 60) / rpm;
}

/* Provide chip's supply voltage, in millivolts */
int
dbcool_supply_voltage(struct dbcool_softc *sc)
{
	if (sc->sc_chip->flags & DBCFLAG_MULTI_VCC)
		if (dbcool_readreg(sc, DBCOOL_CONFIG1_REG) & DBCOOL_CFG1_Vcc)
			return 5000;
		else
			return 3300;
	else if (sc->sc_chip->flags & DBCFLAG_MULTI_VCC2)
		if (dbcool_readreg(sc, DBCOOL_CONFIG1A_REG) & DBCOOL_CFG1A_Vcc)
			return 5000;
		else
			return 3300;
	else
		return 3300;
}

int
dbcool_read_voltage(struct dbcool_softc *sc, uint8_t reg)
{
	uint8_t ext, v1, v2, v3, v4, val;
	int ret, nom;

	/* ADT7466 voltages are in strange locations with only 8-bits */
	if (reg == DBCOOL_AIN1 || reg == DBCOOL_AIN2 || reg == DBCOOL_VCCALT) {
		ext = 0;
		val = dbcool_readreg(sc, reg);
		if (reg == DBCOOL_VCCALT)
			nom = 3300;
		else
			nom = 2250;
	} else if (reg == DBCOOL_12VIN) {
		ext = dbcool_readreg(sc, DBCOOL_EXTRES2_REG) && 0x03;
		val = dbcool_readreg(sc, reg);
		nom = 12000;
		(void)dbcool_readtemp(sc, reg);
	} else {
		ext = dbcool_readreg(sc, DBCOOL_EXTRES1_REG);
		v1 = dbcool_readreg(sc, DBCOOL_25VIN);
		v2 = dbcool_readreg(sc, DBCOOL_CPU_VOLTAGE);
		v3 = dbcool_readreg(sc, DBCOOL_SUPPLY_VOLTAGE);
		v4 = dbcool_readreg(sc, DBCOOL_5VIN);
		switch (reg) {
		case DBCOOL_25VIN:
			val = v1;
			nom = 2500;
			break;
		case DBCOOL_CPU_VOLTAGE:
			val = v2;
			nom = 2250;	/* All known chips use 2.25V reference*/
			ext >>= 2;
			break;
		case DBCOOL_SUPPLY_VOLTAGE:
			val = v3;
			nom = dbcool_supply_voltage(sc);
			ext >>= 4;
			break;
		case DBCOOL_5VIN:
			val = v4;
			nom = 5000;
			ext >>= 6;
			break;
		default:
			val = nom = 0;
		}
		ext &= 0x03;
	}

/* 
 * Scale the nominal value by the 10-bit fraction
 * To avoid overflows, the nominal value is specified in millivolts!
 * Returned value is in microvolts.
 */
	ret = (uint16_t)val << 2 | ext;
	ret = (ret * nom) / 0x300;
	ret *= 1000;

	return ret;
}

SYSCTL_SETUP(sysctl_dbcoolsetup, "sysctl dBCool subtree setup")
{
	sysctl_createv(NULL, 0, NULL, NULL,
		       CTLFLAG_PERMANENT,
		       CTLTYPE_NODE, "hw", NULL,
		       NULL, 0, NULL, 0,
		       CTL_HW, CTL_EOL);
}

static int
sysctl_dbcooltmin(SYSCTLFN_ARGS)
{
	struct sysctlnode node = *rnode;
	struct dbcool_softc *sc = (struct dbcool_softc *)node.sysctl_data;
	int reg = 12345, nd = 0;
	const int *np = newp;
	uint8_t offset = dbcool_tempoffset(sc);
	uint8_t chipreg = (uint8_t)(node.sysctl_idata & 0xff);
	
	reg = (uint32_t)dbcool_readreg(sc, chipreg) - offset;

	node.sysctl_idata = reg;
	if (np) {
		/* we're asked to write */	
		nd = *np;
		node.sysctl_data = &reg;
		if (sysctl_lookup(SYSCTLFN_CALL(&node)) == 0) {
			int8_t new_reg;

			/* Get requested new value, and sanity check it */
			new_reg = (int8_t)node.sysctl_idata;
			if (new_reg < 25 || new_reg > 85)
				return EINVAL;

			dbcool_writereg(sc, chipreg, new_reg + offset);
			return 0;
		}
		return EINVAL;
	} else {
		node.sysctl_size = 4;
		return(sysctl_lookup(SYSCTLFN_CALL(&node)));
	}
}

static int
sysctl_dbcoolduty(SYSCTLFN_ARGS)
{
	struct sysctlnode node = *rnode;
	struct dbcool_softc *sc = (struct dbcool_softc *)node.sysctl_data;
	int reg = 12345, nd = 0;
	const int *np = newp;
	uint8_t chipreg = (uint8_t)(node.sysctl_idata & 0xff);
	
	reg = (uint32_t)dbcool_readreg(sc, chipreg);

	node.sysctl_idata = reg * 100 / 255;
	if (np) {
		/* we're asked to write */	
		nd = *np;
		node.sysctl_data = &reg;
		if (sysctl_lookup(SYSCTLFN_CALL(&node)) == 0) {
			int8_t new_reg;

			/* Get requested new value, and sanity check it */
			new_reg = (int8_t)node.sysctl_idata;
			if (new_reg < 0 || new_reg > 100)
				return EINVAL;

			dbcool_writereg(sc, chipreg, new_reg * 255 / 100);
			return 0;
		}
		return EINVAL;
	} else {
		node.sysctl_size = 4;
		return(sysctl_lookup(SYSCTLFN_CALL(&node)));
	}
}

static int
sysctl_dbcoolmux(SYSCTLFN_ARGS)
{
	struct sysctlnode node = *rnode;
	struct dbcool_softc *sc = (struct dbcool_softc *)node.sysctl_data;
	int reg = 12345, nd = 0;
	const int *np = newp;
	uint8_t chipreg = (uint8_t)(node.sysctl_idata & 0xff);
	
	reg = (uint32_t)dbcool_readreg(sc, chipreg);

	node.sysctl_idata = (reg >> 5) & 0x07;
	if (np) {
		/* we're asked to write */	
		nd = *np;
		node.sysctl_data = &reg;
		if (sysctl_lookup(SYSCTLFN_CALL(&node)) == 0) {
			int8_t new_reg;

			/* Get requested new value, and sanity check it */
			new_reg = (int8_t)node.sysctl_idata;
			if (new_reg < 0 || new_reg > 7)
				return EINVAL;

			new_reg = (reg & 0x1f) | (new_reg << 5);
			dbcool_writereg(sc, chipreg, new_reg);
			return 0;
		}
		return EINVAL;
	} else {
		node.sysctl_size = 4;
		return(sysctl_lookup(SYSCTLFN_CALL(&node)));
	}
}

static int
sysctl_dbcoolrange(SYSCTLFN_ARGS)
{
	struct sysctlnode node = *rnode;
	struct dbcool_softc *sc = (struct dbcool_softc *)node.sysctl_data;
	int reg = 12345, nd = 0;
	const int *np = newp;
	uint8_t chipreg = (uint8_t)(node.sysctl_idata & 0xff);
	
	reg = (uint32_t)dbcool_readreg(sc, chipreg);

	node.sysctl_idata = (reg >> 4) & 0x0f;
	if (np) {
		/* we're asked to write */	
		nd = *np;
		node.sysctl_data = &reg;
		if (sysctl_lookup(SYSCTLFN_CALL(&node)) == 0) {
			int8_t new_reg;

			/* Get requested new value, and sanity check it */
			new_reg = (int8_t)node.sysctl_idata;
			if (new_reg < 0 || new_reg > 0x0f)
				return EINVAL;

			new_reg = (reg & 0x0f) | (new_reg << 4);
			dbcool_writereg(sc, chipreg, new_reg);
			return 0;
		}
		return EINVAL;
	} else {
		node.sysctl_size = 4;
		return(sysctl_lookup(SYSCTLFN_CALL(&node)));
	}
}

#define __DBCOOL_SYSCTL(reg, func, desc, ...)			\
	sysctl_createv(NULL, 0, NULL, 				\
	       (const struct sysctlnode **)&node, 		\
	       CTLFLAG_READWRITE | CTLFLAG_OWNDESC | CTLFLAG_IMMEDIATE, \
	       CTLTYPE_INT, name, desc, func, reg,		\
	       NULL, 0, CTL_HW, __VA_ARGS__,			\
	       CTL_CREATE, CTL_EOL)

static void
dbcool_setup(device_t self)
{
	struct dbcool_softc *sc = device_private(self);
	const struct sysctlnode *me = NULL;
	const struct sysctlnode *me2 = NULL;
	struct sysctlnode *node = NULL;
	uint8_t cfg_val;
	int i, ret;
	int error;
	char name[16];

	sc->sc_sme = sysmon_envsys_create();
	    
	ret = sysctl_createv(NULL, 0, NULL, &me,
	       CTLFLAG_READWRITE,
	       CTLTYPE_NODE, device_xname(self), NULL,
	       NULL, 0, NULL, 0,
	       CTL_HW, CTL_CREATE, CTL_EOL);

	/* create sensors / controllers */
	for (i=0; sc->sc_chip->table[i].reg != 0; i++) {
		strlcpy(sc->sc_sensor[i].desc, sc->sc_chip->table[i].desc,
		    sizeof(sc->sc_sensor[i].desc));
		switch (sc->sc_chip->table[i].type) {
		case DBC_TEMP:
			snprintf(name, 16, "temp%d", sc->sc_chip->table[i].seq);
			sc->sc_sensor[i].units = ENVSYS_STEMP;
			break;
		case DBC_VOLT:
			snprintf(name, 16, "volt%d", sc->sc_chip->table[i].seq);
			sc->sc_sensor[i].units = ENVSYS_SVOLTS_DC;
			break;
		case DBC_FAN:
			snprintf(name, 16, "rpm%d", sc->sc_chip->table[i].seq);
			sc->sc_sensor[i].units = ENVSYS_SFANRPM;
			break;
		case DBC_CTL:
			strlcpy(name, sc->sc_chip->table[i].desc, sizeof(name));
			ret = __DBCOOL_SYSCTL(sc->sc_chip->table[i].reg,
					      sysctl_dbcooltmin,
					      sc->sc_sensor[i].desc,
					      me->sysctl_num);
			if (node != NULL)
				node->sysctl_data = sc;
			break;
		default:
			aprint_error_dev(self, "Bad sensor_table idx %d\n", i);
		}
		if  (sc->sc_chip->table[i].type != DBC_CTL) {
			sc->regs[i] = sc->sc_chip->table[i].reg;
			if (sysmon_envsys_sensor_attach(sc->sc_sme,
							&sc->sc_sensor[i]))
				goto out;
		}
	}

	/* If supported, create sysctl tree for fan PWM controllers */
	if (sc->sc_chip->power != NULL)
		for (i = 0; sc->sc_chip->power[i].desc != NULL; i++) {
			snprintf(name, 16, "fan_ctl_%d", i);
			ret = sysctl_createv(NULL, 0, NULL, &me2,
			       CTLFLAG_READWRITE,
			       CTLTYPE_NODE, name, NULL,
			       NULL, 0, NULL, 0,
			       CTL_HW, me->sysctl_num, CTL_CREATE, CTL_EOL);

			if (sc->sc_chip->flags & DBCFLAG_HAS_PWMMUX) {
				strncpy(name, "mux_control", 16);
				ret = __DBCOOL_SYSCTL(sc->sc_chip->power[i].mux,
						      sysctl_dbcoolmux, NULL,
						      me->sysctl_num,
						      me2->sysctl_num);
				if (node != NULL)
					node->sysctl_data = sc;
			}

			strncpy(name, "min_duty_cycle", 16);
			ret = __DBCOOL_SYSCTL(sc->sc_chip->power[i].min,
					      sysctl_dbcoolduty, NULL,
					      me->sysctl_num, me2->sysctl_num);
			if (node != NULL)
				node->sysctl_data = sc;

			if (sc->sc_chip->flags & DBCFLAG_HAS_MAXDUTY) {
				strncpy(name, "max_duty_cycle", 16);
				ret = __DBCOOL_SYSCTL(sc->sc_chip->power[i].max,
						      sysctl_dbcoolduty, NULL,
						      me->sysctl_num,
						      me2->sysctl_num);
				if (node != NULL)
					node->sysctl_data = sc;
			}

			strncpy(name, "cur_duty_cycle", 16);
			ret = __DBCOOL_SYSCTL(sc->sc_chip->power[i].cur,
					      sysctl_dbcoolduty, NULL,
					      me->sysctl_num, me2->sysctl_num);
			if (node != NULL)
				node->sysctl_data = sc;

			strncpy(name, "temp_range_code", 16);
			ret = __DBCOOL_SYSCTL(sc->sc_chip->power[i].range,
					      sysctl_dbcoolrange, NULL,
					      me->sysctl_num, me2->sysctl_num);
			if (node != NULL)
				node->sysctl_data = sc;
		}

	/* Read and rewrite config register to activate device */
	cfg_val = dbcool_readreg(sc, DBCOOL_CONFIG1_REG);
	if (sc->sc_chip->flags & DBCFLAG_NEEDSTART)
		cfg_val |= DBCOOL_CFG1_START;
	dbcool_writereg(sc, DBCOOL_CONFIG1_REG, cfg_val);

	sc->sc_sme->sme_name = device_xname(self);
	sc->sc_sme->sme_cookie = sc;
	sc->sc_sme->sme_refresh = dbcool_refresh;

	if ((error = sysmon_envsys_register(sc->sc_sme)) != 0) {
		aprint_error_dev(self,
		    "unable to register with sysmon (%d)\n", error);
		goto out;
	}
	
	return;

out:
	sysmon_envsys_destroy(sc->sc_sme);
}

#undef __DBCOOL_SYSCTL

void
dbcool_refresh(struct sysmon_envsys *sme, envsys_data_t *edata)
{
	struct dbcool_softc *sc=sme->sme_cookie;
	int i;
	uint8_t reg;
	
	i = edata->sensor;
	reg = sc->regs[i];
	switch (edata->units)
	{
		case ENVSYS_STEMP:
			edata->value_cur = dbcool_readtemp(sc, reg);
			break;
		case ENVSYS_SVOLTS_DC:
			edata->value_cur = dbcool_read_voltage(sc, reg);
			break;
			
		case ENVSYS_SFANRPM:
			edata->value_cur = dbcool_readrpm(sc, reg);
			break;
	}
	edata->state = ENVSYS_SVALID;
}
#endif /* NSYSMON_ENVSYS > 0 */

static int
dbcool_chip_ident(struct dbcool_softc *sc)
{
	/* verify this is a supported dBCool chip */
	uint8_t c_id, d_id;
	int i;

	c_id = dbcool_readreg(sc, DBCOOL_COMPANYID_REG);
	d_id = dbcool_readreg(sc, DBCOOL_DEVICEID_REG);
    
	for (i = 0; chip_table[i].company != 0; i++)
		if ((c_id == chip_table[i].company) &&
		    (d_id == chip_table[i].device))
			return i;

	aprint_error("dbcool_chip_ident: addr 0x%02x c_id 0x%02x d_id 0x%02x"
		     " No match.\n", sc->sc_addr, c_id, d_id);

	return -1;
}  

